#version 150
precision highp float; precision highp int;
//Ethan Alexander Shulman 2023 "Winter Lights" December 18, 2023

in vec3 in_Position;
out vec4 uv;
out float time;

layout(std140) uniform ShaderviewBlock {
	vec4 mouse;
	float screenX,screenY;
	int frame,framerate;
};
layout(std140) uniform MidiBlock {
	vec4 midi[64];//x = note id, y = velocity, z = channel/track, w = time in seconds since note pressed
	int midiCount;
};
/*LMMS midi converter https://www.lynxwave.com/LMMStoMIDI/LMMStoMIDIConverter.html since LMMS midi export is broken.

Launch options for midi config 'Shaderview console art/WinterLights.mid 0 0'

Midi Instrument 0 = NST Choir High
Midi Instrument 1 = VSCO2 Organ
Midi Instrument 2 = VSCO2 Glockenspi*/

//x = choir, y = organ, z = glockenspiel, w = glocken note
out vec4 music;



const vec3 SNOW_SKY = pow(vec3(60,48,50)*1.5/255., vec3(2.2));

#define ANIMATION 1

//#define FOG(rp) (SNOW_SKY*.3)
#define FOG_DENSITY .07

#define SAMPLES 200
#define EPSILON .004
#define NRM_EPSILON (EPSILON*.5)

#define PI 3.141592653589793



//2d rotation matrix
mat2 r2d(float a) {
	float c = cos(a), s = sin(a);
	return mat2(c,s,-s,c);
}

//pseudo random uvec4 hash
uvec4 hash(uvec4 seed) {
    uvec4 h = (0x6A7F8FAAu^seed)*0x01000193u;
    h = ((h.wxyz>>3u)^h^seed.yzwx)*0x01000193u;
    h = ((h.zwxy>>8u)^h^seed.wxyz)*0x01000193u;
    return h^(h>>11u);
}
#define I2F (1./float(0xFFFFFFFFu))

//r2 hash http://extremelearning.com.au/unreasonable-effectiveness-of-quasirandom-sequences/
vec2 quasiRandom(float n) {
	#define R2V 1.32471795724474602596
	return fract(.5+n*(1./vec2(R2V,R2V*R2V)));
}

vec3 hue(float h) {
	return clamp(abs(mod(h*6.+vec3(0.,4.,2.),6.)-3.)-1.,0.,1.);
}



//geometry distance/material function
vec2 geometry(vec3 rp) {
	//ground snow
	float gd = rp.y-.7+cos(dot(rp.xz,vec2(.06,.4)))*.8 + sin(dot(rp.xz,vec2(1.3,-.23)))*.2 + cos(dot(rp.xz,vec2(.5,1.6)))*.1;
	vec2 geo = vec2(gd,0);
	
	//falling snowflakes
	vec3 mp = rp+10.;
	mp += cos(mp.yzx*.8)*.5;
	mp += time*vec3(.1, 1, .07237);
	
	vec3 sp = mod(mp,2.)-1.;
	vec3 fp = floor(mp/2.);
	sp.xz *= r2d(dot(fp,vec3(.17632, .443875, .2743345)));
	float sd = max(length(sp.xy)-.04, abs(sp.z)-.005);
	if (sd < geo.x) geo = vec2(sd,0);
	
	if (rp.y < 1.) {
		rp += cos((rp.yxz+1.)*vec3(.8,1.9,1.2))*.1;
		
		//string
		float sd = length(rp.yz)-.02;
		if (sd < geo.x) geo = vec2(sd, 1);
	
		//lights
		float xi = floor(rp.x/1.4+10.);
		vec2 rrnd = fract(xi*vec2(.4784,.333573))-.5;
		rp.x = mod(rp.x+14.,1.4)-.7;
		rp.yz *= r2d(rrnd.y);
		rp.xy *= r2d(rrnd.x*1.5);
		rp.y -= .2;
		rp.x = abs(rp.x)+.08;
		float ld = length(rp)-.2;
		if (ld < geo.x) geo = vec2(ld, 2.+xi);
	}
	
	return geo;
}

float detailGeometry(vec3 rp) {
	vec2 gd = geometry(rp);
	if (gd.y == 0.) {
		//snow crystals
		vec3 mp = rp+100.;
		mp += cos(rp.zxy*63.4)*0.01 + sin(rp.yzx*18.12876)*.02;
		vec3 fp = floor(mp/.04);
		mp = mod(mp,.04)-.02;
		mp.xz *= r2d(dot(fp,vec3(1.437864,.327643,2.1734)));
		mp.yz *= r2d(dot(fp,vec3(.17684,3.77824,1.3587324)));
		mp.y *= 10.;
		gd.x = max(gd.x, mp.y*mp.y);
	}
	return gd.x;
}

//returns mat4x3(vec3 diffuse, vec3 specular, vec3 specularData(x=amount, y=roughness, z=fresnel), vec3 emission)
mat4x3 material(vec3 rp, vec3 nrm, float mat) {
	if (mat == 1.) {
		//string
		return mat4x3(vec3(.1,.18,.08),vec3(.9),vec3(.5,.04,.15), vec3(0));
	}
	
	if (mat > 1.5) {
		//light
		mat -= 2.;
		vec3 lcol = hue(mat/6.);
		
		mat += floor(time);
		float light = 20.;
		#if ANIMATION
		if (mod(mat,2.) < .5) {
			//glockenspiel
			light *= music.z*(abs(mod(floor(mat/2.),3.)-music.w)<.5 ? 1. : 0.);
		
		} else {
			if (mod(floor(mat/2.),2.) < .5) {
				//choir
				light *= music.x;
			} else {
				//organ
				light *= music.y;
			}
		}
		#endif
		return mat4x3(lcol*.3,vec3(.95),vec3(.5,.0004,0), lcol*light);
	}


	//ground snow
	return mat4x3(vec3(.7,.77,.85),vec3(.9),vec3(.1,.005,-1), vec3(0));
}

//calculate normal from geometry distance difference
vec3 normal(vec3 rp, float bdst) {
	vec3 nrm = normalize(vec3(geometry(rp+vec3(NRM_EPSILON,0,0)).x, geometry(rp+vec3(0,NRM_EPSILON,0)).x, geometry(rp+vec3(0,0,NRM_EPSILON)).x)-bdst);
	return nrm==nrm ? nrm : vec3(0,1,0);
}

vec3 detailNormal(vec3 rp) {
	vec3 nrm = normalize(vec3(detailGeometry(rp+vec3(NRM_EPSILON,0,0)), detailGeometry(rp+vec3(0,NRM_EPSILON,0)), detailGeometry(rp+vec3(0,0,NRM_EPSILON)))-detailGeometry(rp));
	return nrm==nrm ? nrm : vec3(0,1,0);
}


//return background color from ray
vec3 background(vec3 rp, vec3 rd) {
	return SNOW_SKY*mix(vec3(.7,.77,.85)*.6, vec3(1), max(rd.y,0.));
}
vec3 backgroundLighting(vec3 rp, vec3 rd) {
	return background(rp,rd);
}


uniform sampler2D Buffer0;

void vert() {
	#ifdef ANIMATION
	time = float(frame)/float(framerate*SAMPLES);
	
	music = vec4(0);
	for (int i = 0; i < midiCount; i++) {
		switch (int(midi[i].z)) {
			case 0://choir
				music.x = 1.;
				break;
		
			case 1://organ
				music.y = 1.;
				break;
				
			case 2://glockenspiel
				music.z = 1.;
				music.w = mod(midi[i].x, 3.);
				break;
		}
	}
	#else
	
	time = 23.;
	
	#endif
	
	//anti aliasing jitter
	uv = vec4((in_Position.xy*vec2(screenX/screenY,1)+(quasiRandom(float(frame%SAMPLES))-.5)*2./screenY),
		(in_Position.xy*.5+.5)*vec2(screenX,screenY));
	gl_Position = vec4(in_Position,1);
}
void frag() {
	//setup camera origin position and direction
	vec3 rp = vec3(0,4.,-6.),
		rd = normalize(vec3(uv.xy,2));
	rd.yz *= r2d(-.6);
	
	//path tracing
	bool hit = false;
	vec3 color = vec3(1), light = vec3(0);
	#ifdef FOG
	vec3 initRp = rp, hitRp;
	#endif
	for (int i = 0; i < 120 && rp.y < 4.1; i++) {
		vec2 geo = geometry(rp);
		if (geo.x < EPSILON) {
			//hit get object/material data
			vec3 nrm = detailNormal(rp);
			mat4x3 mat = material(rp,nrm,geo.y);
			vec4 rnd = hash(uvec4(abs(rp)*10000.,frame%SAMPLES))*I2F;
			
			//fog
			#ifdef FOG
			if (!hit) hitRp = rp;
			#endif
			hit = true;
			
			//object emission
			light += color*mat[3];
			
			//choose diffuse/specular ray
			if (rnd.w >= mat[2].x) {
				//diffuse(uniformly distributed ray direction around hemisphere of normal)
				color *= mat[0];
				
				float cs = rnd.x*2.-1., sn = sqrt(1.-cs*cs), ang = rnd.y*PI*2.;
				rd = normalize(vec3(cs,sn*cos(ang),sn*sin(ang)));//random ray direction
				if (dot(rd,nrm) < 0.) rd = -rd;//limit bounce to normal hemisphere
			
			} else {
				//specular(reflection off exponential distributed direction)
				float fresnel = clamp((1.+dot(rd,nrm)-mat[2].z)/max(1e-6,1.-mat[2].z),0.,1.);
				color *= mat[1]*fresnel;
				
				vec3 tang = normalize(cross(abs(nrm.y)>(1.-1e-5)?vec3(0,0,-1):vec3(0,-1,0),nrm)), bitang = cross(tang,nrm);//tangent and binormal
				float cs = pow(rnd.x,mat[2].y), sn = sqrt(1.-cs*cs), ang = rnd.y*PI*2.;
				rd = normalize(reflect(rd, mat3(nrm,tang,bitang)*normalize(vec3(cs,sn*cos(ang),sn*sin(ang)))));//reflect off random direction(simulate microsurface bumps)
			}
			
			//step ray back out of objects
			rp += normal(rp,geo.x)*(EPSILON*2.-geo.x);
			
		} else {
			//step ray forward
			rp += rd*geo.x;
		}		
	}
	
	//apply background
	light += hit ? color*backgroundLighting(rp,rd) : background(rp,rd);
	
	//apply fog
	#ifdef FOG
	float fi = 1./exp(length(hitRp-initRp)*FOG_DENSITY);
	light = light*fi+FOG(hitRp)*(1.-fi);
	#endif
	
	//add sample to buffer
	vec4 sum = texelFetch(Buffer0,ivec2(uv.zw),0);
	if (frame%SAMPLES == 0) sum = vec4(0);
	sum += vec4(light, 1.);
	gl_FragColor = sum;
}
